#
# BTZen - library to asynchronously access Bluetooth devices.
#
# Copyright (C) 2015-2020 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Connection manager to manage connections of multiple Bluetooth devices.

Based on

    https://gist.github.com/parthitce/eb6b751df3235f7247babc4c9aba41d8

NOTE: Discovery is still required on devices like Raspberry Pi to reconnect
    a long running device.

When starting and running connection manager

1. Register Bluetooth agent.
2. Start discovery (it seems to be still required on devices like Raspberry
   Pi in order to reconnect disconnected devices).
3. Create connection manager object on D-Bus event bus and register UUIDs
   of services of devices to be managed by the connection manager.
4. For each device
4.1. Use `ConnectDevice` method of adapter interface to connect to the
     device.
4.2. Wait for `ServicesResolved` property to be changed.
4.3.1. If the property is set to true enable Bluetooth device.
4.3.2. If the property is set to false, then disable Bluetooth device.

When connection manager is closed

1. Close each Bluetooth device.
2. Disconnect each Bluetooth device.
3. Remove each Bluetooth device.
4. Unregister Bluetooth agent.
5. Close connection manager.

"""

import asyncio
import logging
from collections import defaultdict
from functools import partial
from itertools import chain
from operator import attrgetter

from .bus import Bus
from . import _cm

flatten = chain.from_iterable

logger = logging.getLogger(__name__)

FMT_PATH_ADAPTER = '/org/bluez/{}'.format

class ConnectionManager:
    def __init__(self, interface='hci0'):
        self._interface = interface
        self._devices = defaultdict(set)
        self._connected = {}
        self._process = False
        self._handle = None

        self.enable_timeout = 30

    def add(self, *devices):
        for dev in devices:
            self._devices[dev.mac].add(dev)
            dev._cm = self
            dev._bus = self._get_bus()
        for mac in self._devices:
            self._connected[mac] = asyncio.Event()

    def close(self):
        """
        Close connection manager.

        All managed devices are closed and disconnected.
        """
        bus = self._get_bus()
        self._process = False
        for dev in flatten(self._devices.values()):
            dev.close()

        adapter_path = FMT_PATH_ADAPTER(self._interface)
        for mac in self._devices:
            dev_path = bus.dev_path(mac)

            _cm.bt_disconnect(bus.system_bus, dev_path)
            logger.info('device {} disconnected'.format(mac))

            _cm.bt_remove(bus.system_bus, adapter_path, dev_path)
            logger.info('device {} removed'.format(mac))

        _cm.bt_unregister_agent(bus.system_bus)

        if self._handle is not None:
            _cm.cm_close(bus.system_bus, adapter_path, self._handle)
        logger.info('connection manager closed')

    async def connected(self, mac):
        if not mac in self._connected:
            raise ValueError(
                'Device with address {} not managed by the connection manager'
                .format(mac)
            )
        await self._connected[mac].wait()

    def __await__(self):
        self._process = True
        bus = self._get_bus()
        path = FMT_PATH_ADAPTER(self._interface)

        yield from _cm.bt_register_agent(bus.system_bus).__await__()
        yield from _cm.bt_start_discovery(bus.system_bus, path).__await__()

        # TODO: if bluez daemon is restarted, the connection manager needs
        # to be reinitialized
        handle = yield from _cm.cm_init(bus.system_bus, path, self).__await__()
        self._handle = handle

        f = self._reconnect
        tasks = (f(mac, devs) for mac, devs in self._devices.items())
        yield from asyncio.gather(*tasks).__await__()

    async def _reconnect(self, mac, devices):
        bus = self._get_bus()

        # determine connection address type; this is a hack, we need better
        # solution in the future; favour random address type; this is
        # useful when battery level is used (specifies no address type, so
        # defaults to "public) and thingy52 sensors (they require "random"
        # address type)
        address_types = set(
            getattr(dev, 'ADDRESS_TYPE', 'public') for dev in devices
        )
        address_type = 'random' if 'random' in address_types else 'public'

        # enable monitoring of the `ServicesResolved` property first
        bus._dev_property_start(mac, 'ServicesResolved')

        adapter_path = FMT_PATH_ADAPTER(self._interface)

        # remove a device preemptively; if it is left in bluez daemon
        # registry, then it might interfere with establishing of new
        # connection
        await self._preempt_remove_dev(bus, mac, adapter_path)
        await self._connect_dev(bus, mac, adapter_path, address_type)

        try:
            await self._restart(mac, devices)
        finally:
            bus._dev_property_stop(mac, 'ServicesResolved')

    async def _restart(self, mac, devices):
        """
        Re-enable or hold Bluetooth device when property 'ServicesResolved`
        changes.
        """
        bus = self._get_bus()
        enable = partial(self._enable, mac, devices)
        cn_clear = self._connected[mac].clear
        cn_set = self._connected[mac].set

        # the `ServicesResolved` property monitoring is started by a
        # caller, so just wait for the service to be resolved
        async for _ in self._resolve_services(mac):
            try:
                logger.info('enabling device {}'.format(mac))
                await enable()
            except Exception as ex:
                logger.info(
                    'enabling device %s failed, seems to be not connected',
                    mac
                )
                if __debug__:
                    logger.exception('error when enabling %s', mac)
                cn_clear()
            else:
                cn_set()

    async def _resolve_services(self, mac):
        """
        Asynchronous generator waiting for a Bluetooth device to be
        resolved.
        """
        bus = self._get_bus()
        cn_clear = self._connected[mac].clear
        while self._process:
            logger.info(
                'device {} waiting for services resolved status change'
                .format(mac)
            )
            resolved = await bus._dev_property_get(mac, 'ServicesResolved')
            logger.info('device {} services resolved: {}'.format(mac, resolved))

            if resolved:
                yield
            else:
                cn_clear()

    async def _enable(self, mac, devices):
        # NOTE: some devices might deadlock when trying to enable multiple
        # subsystems, i.e. sensor tag with button and temperature only; use
        # timeout to try to avoid the deadlock
        while self._process:
            timeout = self.enable_timeout
            try:
                tasks = (dev.enable() for dev in devices)
                tasks = (asyncio.wait_for(t, timeout=timeout) for t in tasks)
                # serialize enabling of devices as the devices might use
                # the common bluetooth characteristics for configuration,
                # i.e. thingy:52
                for t in tasks:
                    await t
                break
            except asyncio.TimeoutError as ex:
                logger.exception('enabling device %s failed due to timeout', mac)

    async def _connect_dev(self, bus, mac, adapter_path, address_type):
        # connect to a device
        #
        # NOTE: bluez 5.50 - scanning by external programs shall be off or
        # we will never connect
        try:
            logger.info(
                'connect device {} via controller {}, address type {}'
                .format(mac, adapter_path, address_type)
            )
            await _cm.bt_connect(bus.system_bus, adapter_path, mac, address_type)
        except Exception as ex:
            if str(ex) == 'Already Exists':
                logger.info('connection for {} already exists'.format(mac))
            else:
                bus._dev_property_stop(mac, 'ServicesResolved')
                raise

    async def _preempt_remove_dev(self, bus, mac, adapter_path):
        dev_path = bus.dev_path(mac)
        try:
            _cm.bt_remove(bus.system_bus, adapter_path, dev_path)
        except:
            if __debug__:
                logger.exception(
                    'preemptive removal of device {} failed, moving on'
                    .format(mac)
                )

    def _get_bus(self):
        return Bus.get_bus(self._interface)

# vim: sw=4:et:ai



#
# BTZen - library to asynchronously access Bluetooth devices.
#
# Copyright (C) 2015-2020 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from functools import wraps, partial
from inspect import iscoroutinefunction

to_int = partial(int.from_bytes, byteorder='little')

#
# from https://github.com/agronholm/asyncio_extras/blob/master/asyncio_extras/contextmanager.py
# MIT license
#
class _AsyncContextManager:
    __slots__ = 'generator'

    def __init__(self, generator) -> None:
        self.generator = generator

    def __aenter__(self):
        return self.generator.asend(None)

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if exc_val is not None:
            await self.generator.athrow(exc_val)
        else:
            try:
                await self.generator.asend(None)
            except StopAsyncIteration:
                pass
            else:
                raise RuntimeError("async generator didn't stop")

def contextmanager(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        generator = func(*args, **kwargs)
        return _AsyncContextManager(generator)

    return wrapper

# vim: sw=4:et:ai



#
# BTZen - library to asynchronously access Bluetooth devices.
#
# Copyright (C) 2015-2020 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Driver for serial device over Bluetooth Smart connection. Implements
Stollmann protcol

    http://www.telit.com/fileadmin/user_upload/products/Downloads/sr-rf/BlueMod/TIO_Implementation_Guide_r04.pdf

Tested with HeinrichsWeikamp OSTC 2 dive computer.
"""

import asyncio
import math
import logging
from binascii import hexlify
from functools import partial

from . import _btzen
from .bus import Bus
from .device import Device, Info, to_uuid
from .util import contextmanager

logger = logging.getLogger(__name__)

def credits_for(n):
    """
    Calculate number of required credits to send `n` number of bytes.
    """
    return min(255, math.ceil(n / 20))

class Serial(Device):
    info = Info(to_uuid(0xfefb))

    UUID_RX_UART = '00000001-0000-1000-8000-008025000000'
    UUID_TX_UART = '00000002-0000-1000-8000-008025000000'
    UUID_TX_CREDIT = '00000004-0000-1000-8000-008025000000'
    UUID_RX_CREDIT = '00000003-0000-1000-8000-008025000000'

    def __init__(self, mac):
        super().__init__(mac)
        self._buffer = bytearray()
        self._init_paths()

    async def enable(self):
        get_path = partial(self._bus.characteristic_path, self.mac)

        self._tx_credit_path = get_path(self.UUID_TX_CREDIT)
        self._tx_uart_path = get_path(self.UUID_TX_UART)
        self._rx_credit_path = get_path(self.UUID_RX_CREDIT)
        self._rx_uart_path = get_path(self.UUID_RX_UART)

        self._bus._gatt_start(self._tx_credit_path)
        self._bus._gatt_start(self._tx_uart_path)

        self._tx_credit = partial(self._bus._gatt_get, self._tx_credit_path)
        self._tx_uart = partial(self._bus._gatt_get, self._tx_uart_path)
        self._tx_credit_size = partial(self._bus._gatt_size, self._tx_credit_path)

        self._rx_credits = 0
        await self._add_rx_credits()
        logger.debug('requesting tx credits on enable')
        value = await self._tx_credit()
        logger.debug('got tx credits on enable: {}'.format(value))

    async def read(self, n):
        await self._cm.connected(self.mac)
        data = bytearray(self._buffer)
        while len(data) < n:
            async with self._rx_credits_mgr(n - len(data)):
                item = await self._tx_uart()
                data.extend(item)

                if __debug__:
                    logger.debug(
                        'bytes read {}, last {}, tx credits size {}'
                        .format(len(data), hexlify(data[-5:]), self._tx_credit_size())
                    )

        assert len(data) >= n

        # keep extra data in buffer, return only requested number of bytes
        self._buffer = data[n:]
        return data[:n]

    async def write(self, data):
        await self._cm.connected(self.mac)
        assert len(data) <= 20

        if self._rx_credits < 1:
            await self._add_rx_credits()

        if self._tx_credit_size() > 0:
            logger.debug('requesting tx credits')
            value = await self._tx_credit()
            logger.debug('got tx credits: {}'.format(value))

        await self._write(self._rx_uart_path, data)

    def close(self):
        """
        Close serial device.
        """
        if self._tx_credit_path:
            self._bus._gatt_stop(self._tx_credit_path)
        if self._tx_uart_path:
            self._bus._gatt_stop(self._tx_uart_path)
        self._init_paths()

    @contextmanager
    async def _rx_credits_mgr(self, n):
        if self._rx_credits < 1:
            await self._add_rx_credits(credits_for(n))
        try:
            yield
        finally:
            self._rx_credits -= 1

    async def _add_rx_credits(self, n=0x20):
        await self._write(self._rx_credit_path, bytes([n]))
        self._rx_credits += n
        logger.debug('rx credits: {}'.format(self._rx_credits))

    async def _write(self, path, data):
        task = _btzen.bt_write(self._bus.system_bus, path, data)
        await task

    def _init_paths(self):
        self._tx_credit_path = None
        self._tx_uart_path = None
        self._rx_credit_path = None
        self._rx_uart_path = None

# vim: sw=4:et:ai



#
# BTZen - library to asynchronously access Bluetooth devices.
#
# Copyright (C) 2015-2020 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Nordic Thingy:52 Bluetooth device sensors

    https://nordicsemiconductor.github.io/Nordic-Thingy52-FW/documentation/firmware_architecture.html

The sensors do not implement the Bluetooth Environmental Sensing interfaces
and require custom classes. All Thingy:52 Bluetooth device sensors are
notifying.
"""

import asyncio
import logging
import struct
import typing as tp
from dataclasses import dataclass, replace

from . import _btzen
from .device import InfoCharacteristic, InfoEnvSensing, DeviceEnvSensing, \
    DeviceCharacteristic, Trigger, TriggerCondition
from .util import to_int

logger = logging.getLogger(__name__)

CONFIG_DATA_FMT = struct.Struct('<HHHHBBBB')

# function to convert 16-bit UUID to full 128-bit Thingy:52 UUID
to_uuid = 'ef68{:04x}-9b35-4933-9b10-52ffa9740042'.format

@dataclass
class Config:
    """
    Thingy:52 Bluetooth device configuration for the weather service
    sensors.
    """
    # temperature sensor data read interval
    temperature: float = 1.0
    # pressure sensor data read interval
    pressure: float = 1.0
    # humidity sensor data read interval
    humidity: float = 1.0
    # color sensor data read interval
    color: float = 1.0
    # gas sensor data read interval (1 - 1s, 2 - 10s, 3 - 60s)
    gas: int = 1
    # color sensor LED calibration (RGB value)
    rgb: tp.Tuple[int, int, int] = (0, 255, 0)

class DeviceThingy52EnvSensing(DeviceEnvSensing):
    """
    Thingy:52 Bluetooth device sensor.
    """
    ADDRESS_TYPE = 'random'

    # NOTE: the configuration is shared by all sensors per given device
    # (mac address)
    CONFIG = {}

    def __init__(self, mac, notifying=True):
        super().__init__(mac, notifying=notifying)

        DeviceThingy52EnvSensing.CONFIG[mac] = Config()
        self.set_trigger(Trigger(TriggerCondition.FIXED_TIME, 1))

    def _trigger_data(self, trigger: Trigger) -> bytes:
        assert trigger.condition == TriggerCondition.FIXED_TIME

        # replace configuration for given thingy52 device
        mac = self.mac
        data = {self.config_attr: trigger.operand}
        config = replace(DeviceThingy52EnvSensing.CONFIG[self.mac], **data)
        DeviceThingy52EnvSensing.CONFIG[mac] = config
        logger.info('thingy52 configuration: {}'.format(config))

        to_ms = lambda v: int(v * 1000)
        data = CONFIG_DATA_FMT.pack(
            to_ms(config.temperature),
            to_ms(config.pressure),
            to_ms(config.humidity),
            to_ms(config.color),
            config.gas,
            *config.rgb,
        )
        return data

class Temperature(DeviceThingy52EnvSensing):
    """
    Thingy:52 Bluetooth device temperature sensor.
    """
    info = InfoEnvSensing(
        to_uuid(0x0200),
        to_uuid(0x0201),
        2,
        uuid_trigger=to_uuid(0x0206),
    )
    config_attr = 'temperature'

    def get_value(self, data):
        return data[0] + data[1] / 100

class Pressure(DeviceThingy52EnvSensing):
    """
    Thingy:52 Bluetooth device pressure sensor.
    """
    info = InfoEnvSensing(
        to_uuid(0x0200),
        to_uuid(0x0202),
        5,
        uuid_trigger=to_uuid(0x0206),
    )
    config_attr = 'pressure'

    def get_value(self, data):
        return to_int(data[:4]) * 100 + data[4]

class Humidity(DeviceThingy52EnvSensing):
    """
    Thingy:52 Bluetooth device humidity sensor.
    """
    info = InfoEnvSensing(
        to_uuid(0x0200),
        to_uuid(0x0203),
        1,
        uuid_trigger=to_uuid(0x0206),
    )
    config_attr = 'humidity'

    def get_value(self, data):
        return data[0]

class DeviceThingy52Configuration(DeviceCharacteristic):
    """
    Thingy:52 Bluetooth device configuration.
    """
    ADDRESS_TYPE = 'random'

class ConnectionParameters(DeviceThingy52Configuration):
    info = InfoCharacteristic(
        to_uuid(0x0100),
        to_uuid(0x0104),
        8,
    )
    async def set_params(
        self,
        min_conn_interval,
        max_conn_interval,
        slave_latency,
        supervision_timeout,
    ):
        system_bus = self._bus.system_bus
        data = struct.pack(
            '<HHHH',
            min_conn_interval,
            max_conn_interval,
            slave_latency,
            supervision_timeout,
        )
        await _btzen.bt_write(system_bus, self._path_data, data)

# vim: sw=4:et:ai



#
# BTZen - library to asynchronously access Bluetooth devices.
#
# Copyright (C) 2015-2020 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import asyncio
import contextvars
import logging
from functools import partial

from . import _btzen
from . import _sd_bus
from .error import ConnectionError

logger = logging.getLogger(__name__)

INTERFACE_DEVICE = 'org.bluez.Device1'
INTERFACE_GATT_CHR = 'org.bluez.GattCharacteristic1'

def _mac_to_path(mac):
    return mac.replace(':', '_').upper()

class Bus:
    bus = contextvars.ContextVar('bus', default=None)

    def __init__(self, system_bus, interface):
        self.system_bus = system_bus
        self.interface = interface

        loop = asyncio.get_event_loop()
        process = partial(_btzen.bt_process, system_bus)
        loop.add_reader(system_bus.fileno, process)

        self._notifications = Notifications(self)

    @staticmethod
    def get_bus(interface: str) -> 'Bus':
        """
        Get system bus reference.

        The reference is local to current thread.

        :param interface: Bluetooth interface, i.e. `hci0`.
        """
        bus = Bus.bus.get()
        if bus is None:
            system_bus = _sd_bus.default_bus()
            bus = Bus(system_bus, interface)
            Bus.bus.set(bus)
        return bus

    def characteristic_path(self, mac, uuid):
        prefix = self.dev_path(mac)
        return _btzen.bt_characteristic(self.system_bus, prefix, uuid)

    def _gatt_start(self, path):
        # TODO: creates notification session; if another session started,
        # then we get notifications twice; this needs to be fixed
        self._notifications.start(path, INTERFACE_GATT_CHR, 'Value')
        _btzen.bt_notify_start(self.system_bus, path)

    def dev_path(self, mac):
        return '/org/bluez/{}/dev_{}'.format(self.interface, _mac_to_path(mac))

    async def _gatt_get(self, path, timeout=0):
        task = self._notifications.get(path, INTERFACE_GATT_CHR, 'Value')
        return (await task)

    def _gatt_stop(self, path):
        _btzen.bt_notify_stop(self.system_bus, path)
        self._notifications.stop(path, INTERFACE_GATT_CHR)

    def _gatt_size(self, path) -> int:
        return self._notifications.size(path, INTERFACE_GATT_CHR, 'Value')

    def _dev_property_start(self, mac, name, iface=INTERFACE_DEVICE):
        path = self.dev_path(mac)
        self._notifications.start(path, iface, name)

    async def _dev_property_get(
            self,
            mac,
            name,
            iface=INTERFACE_DEVICE,
            timeout=0
        ):
        path = self.dev_path(mac)
        value = await self._notifications.get(path, iface, name)
        return value

    def _dev_property_stop(self, mac, name, iface=INTERFACE_DEVICE):
        path = self.dev_path(mac)
        self._notifications.stop(path, iface)

    async def _property(self, mac, iface, name, type='s', timeout=0):
        bus = self.system_bus
        path = self.dev_path(mac)
        value = await _btzen.bt_property(bus, path, iface, name, type)
        return value

    async def _get_name(self, mac):
        path = self.dev_path(mac)
        bus = self.system_bus
        value = await _btzen.bt_property(bus, path, INTERFACE_DEVICE, 'Name', 's')
        return value

class Notifications:
    def __init__(self, bus):
        self._data = {}
        self._bus = bus

    def start(self, path, iface, name):
        key = path, iface
        data = self._data.get(key)
        if data is None:
            bus = self._bus.system_bus
            data = _btzen.bt_property_monitor_start(bus, path, iface)
            self._data[key] = data

        assert key in self._data
        if not data.is_registered(name):
            data.register(name)

    def size(self, path, iface, name):
        key = path, iface
        return self._data[key].size(name)

    async def get(self, path, iface, name):
        key = path, iface
        data = self._data[key]
        return (await data.get(name))

    def stop(self, path, iface):
        # TODO: add name and call PropertyNotification.stop when no
        # properties monitored
        key = path, iface
        data = self._data[key].stop()

# vim: sw=4:et:ai



#
# BTZen - library to asynchronously access Bluetooth devices.
#
# Copyright (C) 2015-2020 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Texas Instrument Sensor Tag Bluetooth device sensors.

The sensors do not implement the Bluetooth Environmental Sensing interfaces
and require custom classes. To enable reading data from sensors, trigger is
required, which we default to one second. Only one trigger can be
configured.

The identificators for specific sensors can be found at

CC2541DK
    http://processors.wiki.ti.com/index.php/SensorTag_User_Guide
CC2650STK
    http://processors.wiki.ti.com/index.php/CC2650_SensorTag_User's_Guide
"""

import asyncio
import enum
import logging
import struct

from .device import InfoCharacteristic, InfoEnvSensing, DeviceEnvSensing, \
    DeviceCharacteristic, Trigger, TriggerCondition
from .device import to_uuid as to_bt_uuid
from .util import to_int

logger = logging.getLogger(__name__)

# function to convert 16-bit UUID to full 128-bit Sensor Tag UUID
to_uuid = 'f000{:04x}-0451-4000-b000-000000000000'.format

class DeviceSensorTag(DeviceEnvSensing):
    """
    Sensor Tag Bluetooth device sensor.
    """
    def __init__(self, mac, notifying=False):
        super().__init__(mac, notifying=notifying)

        self.set_trigger(Trigger(TriggerCondition.FIXED_TIME, 1))

    async def _configure(self):
        await super()._configure()
        # allow to read first value from sensor
        await asyncio.sleep(self._trigger.operand)

    def _trigger_data(self, trigger: Trigger) -> bytes:
        assert trigger.condition == TriggerCondition.FIXED_TIME

        value = int(trigger.operand * 100)
        assert value < 256
        return bytes([value])

class Temperature(DeviceSensorTag):
    """
    Sensor Tag Bluetooth device temperature sensor.
    """
    info = InfoEnvSensing(
        to_uuid(0xaa00),
        to_uuid(0xaa01),
        4,
        to_uuid(0xaa02),
        to_uuid(0xaa03),
        b'\x01',
        b'\x01',
        b'\x00',
    )

    def get_value(self, data):
        return to_int(data[2:]) / 128.0

class Pressure(DeviceSensorTag):
    """
    Sensor Tag Bluetooth device pressure sensor.
    """
    info = InfoEnvSensing(
        to_uuid(0xaa40),
        to_uuid(0xaa41),
        6,
        to_uuid(0xaa42),
        to_uuid(0xaa44),
        b'\x01',
        b'\x01',
        b'\x00',
    )

    def get_value(self, data):
        return to_int(data[3:])

class Humidity(DeviceSensorTag):
    """
    Sensor Tag Bluetooth device humidity sensor.
    """
    HDC1000_HUMIDITY = 65536 / 100

    info = InfoEnvSensing(
        to_uuid(0xaa20),
        to_uuid(0xaa21),
        4,
        to_uuid(0xaa22),
        to_uuid(0xaa23),
        b'\x01',
        b'\x01',
        b'\x00',
    )

    def get_value(self, data):
        return to_int(data[2:]) / self.HDC1000_HUMIDITY

class Light(DeviceSensorTag):
    """
    Sensor Tag Bluetooth device light sensor.
    """
    info = InfoEnvSensing(
        to_uuid(0xaa70),
        to_uuid(0xaa71),
        2,
        to_uuid(0xaa72),
        to_uuid(0xaa73),
        b'\x01',
        b'\x01',
        b'\x00',
    )

    def get_value(self, data):
        v = to_int(data)
        m = (v & 0x0FFF) / 100
        e = (v & 0xF000) >> 12
        return m * (2 << e)

class Accelerometer(DeviceSensorTag):
    """
    Sensor Tag Bluetooth device accelerometer sensor.
    """
    ACCEL = 0x08 | 0x10 | 0x20
    WAKE_ON_MOTION = 0x80

    MPU9250_GYRO = 65536 / 500
    MPU9250_ACCEL_2G = 32768 / 2
    MPU9250_ACCEL_UNPACK = struct.Struct('<3h').unpack

    info = InfoEnvSensing(
        to_uuid(0xaa80),
        to_uuid(0xaa81),
        18,
        to_uuid(0xaa82),
        to_uuid(0xaa83),
        struct.pack('<H', ACCEL),
        struct.pack('<H', ACCEL | WAKE_ON_MOTION),
        b'\x00\x00',
    )

    def get_value(self, data):
        # gyro: data[:6]
        # magnet: data[12:]
        return tuple(
            v / self.MPU9250_ACCEL_2G
            for v in self.MPU9250_ACCEL_UNPACK(data[6:12])
        )

class ButtonState(enum.IntFlag):
    """
    Sensor Tag Bluetooth device button state.
    """
    OFF = 0x00
    USER = 0x01
    POWER = 0x02
    REED_RELAY = 0x04

class Button(DeviceCharacteristic):
    """
    Sensor Tag Bluetooth device button.
    """
    info = InfoCharacteristic(to_bt_uuid(0xffe0), to_bt_uuid(0xffe1), 1)

    def get_value(self, data: bytes) -> ButtonState:
        return ButtonState(data[0])

# vim: sw=4:et:ai



#
# BTZen - library to asynchronously access Bluetooth devices.
#
# Copyright (C) 2015-2020 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# distutils: language = c
# cython: c_string_type=unicode, c_string_encoding=utf8, language_level=3str

cdef extern from "<systemd/sd-bus.h>":
    ctypedef struct sd_bus:
        pass

    ctypedef struct sd_bus_message:
        pass

    ctypedef struct sd_bus_slot:
        pass

    ctypedef struct sd_bus_error:
        const char *name
        const char *message

    ctypedef struct sd_bus_vtable:
        pass

    ctypedef int (*sd_bus_message_handler_t)(sd_bus_message*, void*, sd_bus_error*)
    ctypedef int (*sd_bus_property_get_t)(sd_bus*, const char*, const char*, const char *, sd_bus_message*, void*, sd_bus_error*)

    int sd_bus_default_system(sd_bus**)
    int sd_bus_get_fd(sd_bus*)
    int sd_bus_process(sd_bus*, sd_bus_message**)
    int sd_bus_add_object_vtable(sd_bus*, sd_bus_slot**, const char*, const char*, const sd_bus_vtable*, void*)
    int sd_bus_add_object_manager(sd_bus*, sd_bus_slot**, const char*)

    int sd_bus_call_method(sd_bus*, const char*, const char*, const char*, const char*, sd_bus_error*, sd_bus_message**, const char*, ...)
    int sd_bus_call_method_async(sd_bus*, sd_bus_slot**, const char*, const char*, const char*, const char*, sd_bus_message_handler_t, void*, const char*, ...)
    int sd_bus_message_new_method_call(sd_bus*, sd_bus_message**, const char*, const char*, const char*, const char*)
    int sd_bus_call(sd_bus*, sd_bus_message*, long, sd_bus_error*, sd_bus_message**)
    int sd_bus_call_async(sd_bus*, sd_bus_slot**, sd_bus_message*, sd_bus_message_handler_t, void*, long)
    int sd_bus_reply_method_return(sd_bus_message*, const char*)

    int sd_bus_message_append_array(sd_bus_message*, char, const void*, size_t)
    int sd_bus_message_append_basic(sd_bus_message*, char, const void*)
    int sd_bus_message_append(sd_bus_message*, const char*, ...)
    int sd_bus_message_append_strv(sd_bus_message*, char**)
    int sd_bus_message_open_container(sd_bus_message*, char, const char*)
    int sd_bus_message_close_container(sd_bus_message*)

    int sd_bus_get_property(sd_bus*, const char*, const char*, const char*, const char*, sd_bus_error*, sd_bus_message**, const char*)

    const sd_bus_error *sd_bus_message_get_error(sd_bus_message*)
    const char *sd_bus_message_get_path(sd_bus_message*)
    int sd_bus_message_read(sd_bus_message*, const char*, ...)
    int sd_bus_message_read_basic(sd_bus_message*, char, void*)
    int sd_bus_message_read_array(sd_bus_message*, char, const void**, size_t*)
    int sd_bus_message_enter_container(sd_bus_message*, char, const char*)
    int sd_bus_message_exit_container(sd_bus_message*)
    int sd_bus_message_skip(sd_bus_message*, const char*)
    int sd_bus_message_get_type(sd_bus_message*, unsigned char*)
    int sd_bus_message_peek_type(sd_bus_message*, char*, const char**)

    int sd_bus_add_match(sd_bus*, sd_bus_slot**, const char*, sd_bus_message_handler_t, void*)

    sd_bus *sd_bus_unref(sd_bus*)
    sd_bus_message *sd_bus_message_unref(sd_bus_message*)
    void sd_bus_error_free(sd_bus_error*)
    sd_bus_slot* sd_bus_slot_unref(sd_bus_slot*)

# no idea how use these at the moment, see _cm.pyx for workaround
#   sd_bus_vtable SD_BUS_VTABLE_START(uint64_t)
#   sd_bus_vtable SD_BUS_METHOD(const char*, const char*, const char*, sd_bus_message_handler_t, uint64_t)
#   sd_bus_vtable SD_BUS_PROPERTY(const char*, const char*, sd_bus_property_get_t, size_t, uint64_t)
#   sd_bus_vtable SD_BUS_VTABLE_END

cdef sd_bus_error SD_BUS_ERROR_NULL = sd_bus_error(NULL, NULL, 0)

cdef class Bus:
    """
    Python level wrapper around D-Bus connection.
    """
    cdef sd_bus *bus
    cdef readonly int _fd_no

cdef class BusMessage:
    """
    Python level wrapper around D-Bus message structure.
    """
    cdef sd_bus_message *c_obj

# vim: sw=4:et:ai



import sys
import os.path
import sphinx_rtd_theme

sys.path.append(os.path.abspath('.'))
sys.path.append(os.path.abspath('doc'))

import btzen

extensions = [
    'sphinx.ext.autodoc', 'sphinx.ext.autosummary', 'sphinx.ext.doctest',
    'sphinx.ext.todo', 'sphinx.ext.viewcode'
]
project = 'btzen'
source_suffix = '.rst'
master_doc = 'index'

version = release = '0.0.9'
copyright = 'Artur Wroblewski'

epub_basename = 'bme280 - {}'.format(version)
epub_author = 'Artur Wroblewski'

todo_include_todos = True

html_theme = 'sphinx_rtd_theme'
html_static_path = ['static']
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

# vim: sw=4:et:ai



#
# BTZen - library to asynchronously access Bluetooth devices.
#
# Copyright (C) 2015-2020 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Generic access to Bluetooth devices.

See also

    https://www.bluetooth.com/specifications/gatt/characteristics
"""

import enum
import logging
import typing as tp
from dataclasses import dataclass
from functools import partial

from . import _btzen
from .bus import Bus

logger = logging.getLogger(__name__)

# function to convert 16-bit UUID to full 128-bit Bluetooth normative UUID
# string
to_uuid = '0000{:04x}-0000-1000-8000-00805f9b34fb'.format

@dataclass(frozen=True)
class Info:
    """
    Bluetooth device information.

    :var service: UUID of Bluetooth service.
    """
    service: str

@dataclass(frozen=True)
class InfoInterface(Info):
    """
    Bluetooth device information for device providing data via Bluez
    interface property.

    Example of interface for Battery Level Bluetooth characteristic is
    `org.bluez.Battery1` Bluez interface, which provides data via
    `Percentage` property.

    :var interface: Bluez interface name.
    :var property: Property name of the interface.
    :var type: Type of property value.
    """
    interface: str
    property: str
    type: str

@dataclass(frozen=True)
class InfoCharacteristic(Info):
    """
    Bluetooth device information for device providing data via Bluetooth
    characteristic.

    :var uuid_data: UUID of characteristic to read data from.
    :var size: Length of data received from Bluetooth characteristic on
        read.
    """
    uuid_data: str
    size: int

@dataclass(frozen=True)
class InfoEnvSensing(InfoCharacteristic):
    """
    Bluetooth device information for device providing data via Bluetooth
    Environmental Sensing characteristic.

    :var uuid_conf: UUID of characteristic to write and read device
        configuration.
    :var uuid_trigger: UUID of characteristic to write and read device
        trigger data.
    :var config_on: Default configuration of device to switch device on.
    :var config_notify_on: Default configuration of device to switch device
        on in notifying mode. If none, then `config_on` is used.
    :var config_off: Default configuration of device to switch device off.
    """
    uuid_conf: tp.Optional[str] = None
    uuid_trigger: tp.Optional[str] = None
    config_on: tp.Optional[bytes] = None
    config_on_notify: tp.Optional[bytes] = None
    config_off: tp.Optional[bytes] = None

class Device:
    """
    Base class for all Bluetooth devices.

    :var mac: Address of Bluetooth device.
    :var notifying: Indicates if data is read in notifying mode.
    """
    def __init__(self, mac, *, notifying=False):
        self.mac = mac
        self.notifying = notifying

        self._bus = None
        self._cm = None
        self._task = None
        self._read_data = None

    async def enable(self):
        """
        Enable Bluetooth device.

        If a Bluetooth device is connected or reconnected, the enable
        method shall be called to configure/reconfigure the device.
        """
        raise NotImplementedError('Enable method is not implemented')

    async def read(self, timeout=0):
        """
        Read data from Bluetooth device.

        If device is in notifying mode, the data might be returned after
        long period of time. The timeout value is ignored for devices in
        such mode.

        :param timeout: Data read timeout in seconds.
        """
        await self._cm.connected(self.mac)
        self._task = self._read_data(timeout=timeout)
        data = await self._task
        self._task = None
        return self.get_value(data)

    def get_value(self, data):
        """
        Convert data returned by Bluetooth device into value.

        By default no conversion is performed.
        """
        return data

    def close(self):
        """
        Disable device and stop reading its data.

        Pending, asynchronous coroutines are closed.
        """
        task = self._task
        if task is not None:
            task.close()
            self._task = None

        logger.info('device {} closed'.format(self))

    def __repr__(self):
        return '{}/{}'.format(self.mac, self.__class__.__name__)

class DeviceInterface(Device):
    """
    Bluetooth device reading data from Bluez interface property.
    """
    info: InfoInterface

    def __init__(self, mac, *, notifying=False):
        super().__init__(mac, notifying=notifying)

        self._params = {
            'mac': self.mac,
            'name': self.info.property,
            'iface': self.info.interface,
        }
        self._read_data = None

    async def enable(self):
        logger.info('enabling device: {}'.format(self))
        read_notify = partial(self._bus._dev_property_get, **self._params)
        read = partial(self._bus._property, **self._params, type=self.info.type)
        self._read_data = read_notify if self.notifying else read
        if self.notifying:
            self._bus._dev_property_start(**self._params)

    def close(self):
        if self.notifying:
            self._bus._dev_property_stop(**self._params)
        super().close()

class DeviceCharacteristic(Device):
    """
    Bluetooth device reading data from Bluetooth characteristic.
    """
    info: InfoCharacteristic

    def __init__(self, mac, notifying=False):
        super().__init__(mac, notifying=notifying)

        self._path_data = None

    async def enable(self):
        logger.info('enabling device: {}'.format(self))
        notify = self.notifying

        self._path_data = path = self._get_path(self.info.uuid_data)

        read = partial(_btzen.bt_read, self._bus.system_bus, path)
        read_notify = partial(self._bus._gatt_get, path)
        self._read_data = read_notify if notify else read

        await self._configure()

        if notify:
            self._bus._gatt_start(path)

        logger.info('enabled device: {}'.format(self))

    def close(self):
        if self.notifying:
            try:
                self._bus._gatt_stop(self._path_data)
            except Exception as ex:
                logger.warning('cannot stop notifications: {}'.format(ex))

        super().close()

    async def _configure(self):
        return None

    def _get_path(self, uuid):
        return self._bus.characteristic_path(self.mac, uuid)

class DeviceEnvSensing(DeviceCharacteristic):
    """
    Bluetooth device providing data via Bluetooth Environmental Sensing
    characteristic.
    """
    info: InfoEnvSensing

    def __init__(self, mac, notifying=False):
        super().__init__(mac, notifying=notifying)

        self._path_conf = None
        self._path_trigger = None
        self._trigger = None

    def set_trigger(self, trigger: 'Trigger'):
        """
        Set Bluetooth Environmental Sensing device trigger data.

        NOTE: The specification supports up to three triggers.

        :param trigger: Information for first trigger.
        """
        self._trigger = trigger

    def set_interval(self, interval: float) -> None:
        """
        Set fixed time interval for Bluetooth Environmental Sensing device.

        This is equivalent to::

            set_trigger(Trigger(TriggerCondition.FIXED_TIME, interval))

        :param interval: Interval in seconds.
        """
        self.set_trigger(Trigger(TriggerCondition.FIXED_TIME, interval))

    def close(self):
        super().close()

        info = self.info
        system_bus = self._bus.system_bus
        if info.config_off:
            try:
                _btzen.bt_write_sync(system_bus, self._path_conf, info.config_off)
            except Exception as ex:
                logger.warning('cannot switch device off: {}'.format(ex))

    async def _configure(self):
        info = self.info

        self._path_conf = self._get_path(info.uuid_conf)
        self._path_trigger = self._get_path(info.uuid_trigger)

        config_on = info.config_on_notify if self.notifying else info.config_on
        if config_on:
            if __debug__:
                logger.debug(
                    'writing {} configuration data to {}'
                    .format(config_on, self._path_conf)
                )
            await self._write(self._path_conf, config_on)

        path = self._path_trigger
        data = self._trigger_data(self._trigger)
        if path is not None and data is not None:
            await self._write(path, data)
        else:
            logger.warning('setting trigger for {} not supported'.format(self))

    def _trigger_data(self, trigger: 'Trigger') -> bytes:
        """
        Convert Bluetooth Environmental Sensing device trigger information
        into raw data.

        ;param: Trigger information.
        """
        return None

    async def _write(self, path: str, data: bytes) -> None:
        await _btzen.bt_write(self._bus.system_bus, path, data)


class BatteryLevel(DeviceInterface):
    """
    The current charge level of a Bluetooth device battery.
    """
    info = InfoInterface(
        to_uuid(0x180f),
        'org.bluez.Battery1',
        'Percentage',
        'y'
    )

class TriggerCondition(enum.IntEnum):
    """
    Condition value for Bluetooth Environmental Sensing device trigger
    information.

    NOTE: Incomplete, see Bluetooth Environmental Sensing Service
        specification, page 18..
    """
    INACTIVE = 0x00
    FIXED_TIME = 0x01

@dataclass(frozen=True)
class Trigger:
    """
    Bluetooth Environmental Sensing device trigger information.
    """
    condition: TriggerCondition
    operand: tp.Any

# vim: sw=4:et:ai



#
# BTZen - library to asynchronously access Bluetooth devices.
#
# Copyright (C) 2015-2020 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Bluetooth Weight service implementation.
"""

import enum
import struct
import typing as tp
from dataclasses import dataclass

from .device import InfoCharacteristic, DeviceCharacteristic, to_uuid

class WeightFlags(enum.IntFlag):
    IMPERIAL = 0x01
    TIMESTAMP = 0x02
    USER_ID = 0x04
    BMI = 0x08
    RESERVED_1 = 0x10
    RESERVED_2 = 0x20
    RESERVED_3 = 0x40
    RESERVED_4 = 0x80

@dataclass(frozen=True)
class WeightData:
    """
    Weight measurement.

    var flags: Weight scale flags value.
    var weight: Weight measurement value.

    Mi Smart Scale extensions:

    :var stabilized: Indicates if weight stabilized.
    :var load_removed: Indicates if load is removed.
    """
    flags: WeightFlags
    weight: float

    stabilized: tp.Optional[bool] = True
    load_removed: tp.Optional[bool] = False

class WeightMeasurement(DeviceCharacteristic):
    """
    Weight measurement Bluetooth device.
    """
    info = InfoCharacteristic(to_uuid(0x181d), to_uuid(0x2a9d), 9)

    def get_value(self, data: bytes):
        flags, weight = struct.unpack('<BH', data[0:3])
        flags = WeightFlags(flags)

        # TODO: set only when mi smart scale
        stabilized = bool(flags & WeightFlags.RESERVED_2)
        load_removed = bool(flags & WeightFlags.RESERVED_4)

        return WeightData(
            flags,
            weight * 0.005,
            stabilized=stabilized,
            load_removed=load_removed
        )

# vim: sw=4:et:ai



#
# BTZen - library to asynchronously access Bluetooth devices.
#
# Copyright (C) 2015-2020 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# distutils: language = c
# cython: c_string_type=unicode, c_string_encoding=utf8, language_level=3str

from libc.stdio cimport perror
from libc.string cimport strerror
from libc.stdint cimport uint8_t
from libc.errno cimport errno
from cpython.bytes cimport PyBytes_FromStringAndSize

import logging
from contextlib import contextmanager

from ._sd_bus cimport *
from .error import *

logger = logging.getLogger(__name__)

ERROR_HINTS = {
}

cdef class Bus:
    @property
    def fileno(self):
        return self._fd_no

def default_bus():
    """
    Get default system bus connection.
    """
    cdef Bus bus = Bus.__new__(Bus)
    cdef int r

    r = sd_bus_default_system(&bus.bus)
    check_call('connect bus', r)
    bus._fd_no = sd_bus_get_fd(bus.bus)

    return bus

def check_call(msg_err, code):
    """
    Raise call error if a D-Bus call has failed.
    """
    if code < 0:
        msg_err = 'Call failed - {}: {} ({})'.format(
            msg_err, strerror(-code), code
        )
        logger.debug('got error code {}: '.format(code, strerror(-code)))
        raise CallError(msg_err)

#
# sd-bus message parsing
#
class BusMessageError(Error):
    """
    Bus message parsing error.
    """

def check_msg_error(r):
    if r < 0:
        raise BusMessageError(
            'D-Bus message parsing error: {}'.format(strerror(-r))
        )

def task_handle_message(BusMessage bus_msg, task, cls_err, value_type: str):
    """
    Handle asynchronous task result or error.

    :param bus_msg: D-Bus message representation as Python object.
    :param task: Asynchronous task instance.
    :param cls: Class of exception to be raised on error.
    :param value_type: Type of value to be read from D-Bus message.
    """
    cdef const sd_bus_error *error = sd_bus_message_get_error(bus_msg.c_obj)

    if task.done():
        return 0
    elif error and error.message:
        task.set_exception(cls_err(error.message))
    elif value_type is None:
        task.set_result(None)
    else:
        value = msg_read_value(bus_msg, value_type)
        task.set_result(value)
    return 0

@contextmanager
def msg_container(BusMessage bus_msg, str type, str contents):
    """
    Parse single D-Bus container entry of given type and contents.
    """
    cdef char msg_type = ord(type)
    cdef sd_bus_message *msg = bus_msg.c_obj
    assert msg != NULL

    r = sd_bus_message_enter_container(msg, msg_type, contents.encode())
    check_msg_error(r)

    yield

    r = sd_bus_message_exit_container(msg)
    check_msg_error(r)

def msg_container_dict(BusMessage bus_msg, str contents):
    """
    Loop over items of D-Bus message dictionary container.
    """
    with msg_container(bus_msg, 'a', contents):
        for _ in msg_container_loop(bus_msg, 'e', contents[1:-1]):
            yield

def msg_container_loop(BusMessage bus_msg, str type, str contents):
    """
    Loop over items of D-Bus message container.

    For dictionary containers use `msg_container_dict`.
    """
    cdef char msg_type = ord(type)
    cdef sd_bus_message *msg = bus_msg.c_obj

    while True:
        r = sd_bus_message_enter_container(msg, msg_type, contents.encode())
        check_msg_error(r)
        if r == 0:
            break

        yield

        r = sd_bus_message_exit_container(msg)
        check_msg_error(r)

def msg_read_value(BusMessage bus_msg, str type):
    """
    Read a value from a sd-bus message of given type.

    Supported values

    - boolean
    - signed short int
    - string
    - byte array
    - variant
    """
    cdef sd_bus_message *msg = bus_msg.c_obj

    cdef bytes value_str
    cdef int value
    cdef uint8_t value_byte
    cdef signed short value_short
    cdef const void *buff
    cdef size_t buff_size
    cdef char *buff_str
    cdef const char *contents
    cdef char msg_type_v

    msg_type = type.encode()

    if msg_type == b'b':
        r = sd_bus_message_read_basic(msg, 'b', &value)
        check_msg_error(r)
        r_value = value == 1

    elif msg_type == b'n':
        r = sd_bus_message_read_basic(msg, 'n', &value_short)
        check_msg_error(r)
        r_value = value_short

    elif msg_type == b'y':
        r = sd_bus_message_read_basic(msg, 'y', &value_byte)
        check_msg_error(r)
        r_value = value_byte

    elif msg_type == b'ay':
        r = sd_bus_message_read_array(msg, 'y', &buff, &buff_size)
        check_msg_error(r)

        r_value = PyBytes_FromStringAndSize(<char*>buff, buff_size)
        logger.debug('array value of size: {}'.format(buff_size))

    elif msg_type == b's' or msg_type == b'o':
        r = sd_bus_message_read(msg, msg_type, &buff_str)
        check_msg_error(r)
        r_value = <str>buff_str
        logger.debug('string value: {} of size {}'.format(r_value, len(r_value)))

    elif msg_type == b'v':
        r = sd_bus_message_peek_type(msg, &msg_type_v, &contents)
        check_msg_error(r)
        assert chr(msg_type_v) == 'v', (msg_type, contents)

        with msg_container(bus_msg, type, contents):
            r_value = msg_read_value(bus_msg, contents)
    else:
        # FIXME: add support for other types
        raise BusMessageError('Unknown message type: {}'.format(type))

    return r_value

def msg_skip(BusMessage bus_msg, str type):
    """
    Skip D-Bus message entry of given type.
    """
    r = sd_bus_message_skip(bus_msg.c_obj, type.encode())
    check_msg_error(r)

# vim: sw=4:et:ai



#
# BTZen - library to asynchronously access Bluetooth devices.
#
# Copyright (C) 2015-2020 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# distutils: language = c
# cython: c_string_type=unicode, c_string_encoding=utf8, language_level=3str

"""
Bluetooth connection management.
"""

from libc.stdlib cimport malloc, free

import asyncio
import logging
from itertools import chain

from ._sd_bus cimport *
from . import _sd_bus

logger = logging.getLogger(__name__)

flatten = chain.from_iterable

cdef extern from *:
    # cdef sd_bus_vtable *cm_vtable = [
    #     SD_BUS_VTABLE_START(0),
    #     SD_BUS_METHOD('Release', '', '', cm_release, 0),
    #     SD_BUS_PROPERTY('UUIDs', 'as', cm_property, 0, SD_BUS_VTABLE_PROPERTY_CONST),
    #     SD_BUS_VTABLE_END
    # ]
    """
    #include <string.h>
    void create_vtable(sd_bus_message_handler_t f_release, sd_bus_property_get_t f_property, sd_bus_vtable** ret) {
        /* FIXME: fix the local variable issue */
        sd_bus_vtable vtable[] = {
            SD_BUS_VTABLE_START(0),
            SD_BUS_METHOD("Release", "", "", f_release, 0),
            SD_BUS_PROPERTY("UUIDs", "as", f_property, 0, SD_BUS_VTABLE_PROPERTY_CONST),
            SD_BUS_VTABLE_END
        };
        *ret = malloc(4 * sizeof(sd_bus_vtable));
        memcpy(*ret, vtable, 4 * sizeof(sd_bus_vtable));
    }
    """
    void create_vtable(sd_bus_message_handler_t, sd_bus_property_get_t, sd_bus_vtable**)

cdef class ConnectionManagerHandle:
    cdef sd_bus_slot *slot
    cdef sd_bus_vtable *vtable

    def stop(self):
        sd_bus_slot_unref(self.slot)
        free(self.vtable)

cdef int cm_release(sd_bus_message *msg, void *user_data, sd_bus_error *error) with gil:
    return sd_bus_reply_method_return(msg, NULL)

cdef int cm_property(
        sd_bus *bus,
        const char *path,
        const char *interface,
        const char *prop,
        sd_bus_message *reply,
        void *user_data,
        sd_bus_error *error
    ) with gil:

    cdef object cm = <object>user_data
    cdef int r

    uuids = set(dev.info.service.encode() for dev in flatten(cm._devices.values()))
    size = len(uuids) + 1
    cdef char **arr = <char**>malloc(size * sizeof(char*))

    for i, uv in enumerate(uuids):
        arr[i] = uv
    arr[size - 1] = NULL

    r = sd_bus_message_append_strv(reply, arr)
    free(arr)

    _sd_bus.check_call('adding uuids', r)
    return 0

async def cm_init(Bus bus, str path, cm):
    """
    Initialize connection manager.
    """
    cdef sd_bus_slot *slot

    handle = ConnectionManagerHandle()
    create_vtable(cm_release, cm_property, &handle.vtable)

    task = asyncio.get_event_loop().create_future()

    r = sd_bus_add_object_manager(bus.bus, NULL, '/')
    _sd_bus.check_call('add object manager', r)

    r = sd_bus_add_object_vtable(
        bus.bus,
        &slot,
        '/org/btzen/ConnectionManager',
        'org.bluez.GattProfile1',
        handle.vtable,
        <void*>cm
    )
    _sd_bus.check_call('add cm vtable', r)

    r = sd_bus_call_method_async(
        bus.bus,
        NULL,
        'org.bluez',
        path.encode(),
        'org.bluez.GattManager1',
        'RegisterApplication',
        task_cb,
        <void*>task,
         'oa{sv}',
         '/',
         0
    )
    try:
        _sd_bus.check_call('register application call', r)
    except Exception as ex:
        sd_bus_slot_unref(slot)
        raise
    else:
        handle.slot = slot

    await task
    return handle

def cm_close(Bus bus, str path, handle):
    """
    Close connection manager.
    """
    assert bus is not None

    cdef sd_bus_message *msg = NULL
    cdef sd_bus_error error = SD_BUS_ERROR_NULL

    handle.stop()

    r = sd_bus_call_method(
        bus.bus,
        'org.bluez',
        path.encode(),
        'org.bluez.GattManager1',
        'UnregisterApplication',
        &error,
        &msg,
         'o',
         '/',
         NULL
    )

    sd_bus_error_free(&error);
    sd_bus_message_unref(msg);

    _sd_bus.check_call('unregister application call', r)

cdef int task_cb(sd_bus_message *msg, void *user_data, sd_bus_error *ret_error) with gil:
    cdef object task = <object>user_data
    cdef BusMessage bus_msg = BusMessage.__new__(BusMessage)
    bus_msg.c_obj = msg

    return _sd_bus.task_handle_message(bus_msg, task, ConnectionError, None)

async def bt_connect(Bus bus, str path, str address, str address_type):
    """
    Connect to Bluetooth device.

    :param bus: D-Bus reference.
    :param path: D-Bus adapter path.
    :param address: Bluetooth device address.
    :param address_type: Bluetooth device address type (public or random).
    """
    assert bus is not None

    buff_addr = address.encode()
    buff_addr_type = address_type.encode()

    cdef sd_bus_message *msg = NULL
    cdef unsigned char *addr_data = buff_addr
    cdef unsigned char *addr_type_data = buff_addr_type

    task = asyncio.get_event_loop().create_future()
    try:
        r = sd_bus_message_new_method_call(
            bus.bus,
            &msg,
            'org.bluez',
            path.encode(),
            'org.bluez.Adapter1',
            'ConnectDevice'
        )
        _sd_bus.check_call('bt connect call prepare {}'.format(path), r)

        r = sd_bus_message_append(
            msg, 'a{sv}', 2,
            'Address', 's', addr_data,
            "AddressType", "s", addr_type_data,
        )
        _sd_bus.check_call('bt connect call args {}'.format(path), r)

        r = sd_bus_call_async(bus.bus, NULL, msg, task_cb, <void*>task, 0)
        _sd_bus.check_call('bt connect call {}'.format(path), r)

        return (await task)
    finally:
        sd_bus_message_unref(msg)

async def bt_start_discovery(Bus bus, str path):
    """
    Start device discovery.

    :param bus: D-Bus reference.
    :param path: D-Bus adapter path.
    """
    assert bus is not None
    cdef sd_bus_message *msg = NULL

    task = asyncio.get_event_loop().create_future()
    try:
        r = sd_bus_message_new_method_call(
            bus.bus,
            &msg,
            'org.bluez',
            path.encode(),
            'org.bluez.Adapter1',
            'StartDiscovery'
        )
        _sd_bus.check_call('bt start discovery call prepare {}'.format(path), r)

        r = sd_bus_call_async(bus.bus, NULL, msg, task_cb, <void*>task, 0)
        _sd_bus.check_call('bt start discovery call {}'.format(path), r)

        return (await task)
    finally:
        sd_bus_message_unref(msg)

def bt_disconnect(Bus bus, str path):
    """
    Disconnect Bluetooth device.

    :param bus: D-Bus reference.
    :param path: D-Bus device path.
    """
    assert bus is not None

    cdef sd_bus_message *msg = NULL
    cdef sd_bus_error error = SD_BUS_ERROR_NULL

    r = sd_bus_call_method(
        bus.bus,
        'org.bluez',
        path.encode(),
        "org.bluez.Device1",
        'Disconnect',
        &error,
        &msg,
        NULL,
        NULL
    )
    sd_bus_error_free(&error);
    sd_bus_message_unref(msg);

    _sd_bus.check_call('disconnect device', r)

def bt_remove(Bus bus, str adapter, str device):
    """
    Remove Bluetooth device.

    :param bus: D-Bus reference.
    :param adapter: D-Bus adapter path.
    :param device: D-Bus device path.
    """
    assert bus is not None

    cdef sd_bus_message *msg = NULL
    cdef sd_bus_error error = SD_BUS_ERROR_NULL

    buff = device.encode()
    cdef unsigned char *dev_data = buff

    r = sd_bus_call_method(
        bus.bus,
        'org.bluez',
        adapter.encode(),
        "org.bluez.Adapter1",
        'RemoveDevice',
        &error,
        &msg,
        'o',
        dev_data
    )
    sd_bus_error_free(&error);
    sd_bus_message_unref(msg);

    _sd_bus.check_call('remove device', r)

async def bt_register_agent(Bus bus):
    """
    Register pairing agent for BTZen library.

    :param bus: D-Bus reference.
    """
    assert bus is not None
    await _bt_agent_register_agent(bus)
    await _bt_agent_request_default_agent(bus)

async def _bt_agent_register_agent(Bus bus):
    """
    Register auto pair Bluetooth agent.

    :param bus: D-Bus reference.
    """
    assert bus is not None
    cdef sd_bus_message *msg = NULL

    try:
        task = asyncio.get_event_loop().create_future()
        r = sd_bus_message_new_method_call(
            bus.bus,
            &msg,
            'org.bluez',
            '/org/bluez',
            'org.bluez.AgentManager1',
            'RegisterAgent'
        )
        _sd_bus.check_call('agent registration call prepare', r)

        r = sd_bus_message_append(msg, 'os', '/org/btzen/Agent', 'NoInputNoOutput')
        _sd_bus.check_call('agent registration params', r)

        r = sd_bus_call_async(bus.bus, NULL, msg, task_cb, <void*>task, 0)
        _sd_bus.check_call('agent registration call', r)

        return (await task)
    finally:
        sd_bus_message_unref(msg)

async def _bt_agent_request_default_agent(Bus bus):
    """
    Request Bluetooth default agent.

    :param bus: D-Bus reference.
    """
    assert bus is not None
    cdef sd_bus_message *msg = NULL
    cdef unsigned char *arg_data

    try:
        task = asyncio.get_event_loop().create_future()
        r = sd_bus_message_new_method_call(
            bus.bus,
            &msg,
            'org.bluez',
            '/org/bluez',
            'org.bluez.AgentManager1',
            'RequestDefaultAgent'
        )
        _sd_bus.check_call('agent registration call prepare', r)

        r = sd_bus_message_append(msg, 'o', '/org/btzen/Agent')
        _sd_bus.check_call('agent registration params', r)

        r = sd_bus_call_async(bus.bus, NULL, msg, task_cb, <void*>task, 0)
        _sd_bus.check_call('agent registration call', r)

        return (await task)
    finally:
        sd_bus_message_unref(msg)

def bt_unregister_agent(Bus bus):
    """
    Unregister auto pair agent.

    :param bus: D-Bus reference.
    """
    assert bus is not None

    cdef sd_bus_message *msg = NULL
    cdef sd_bus_error error = SD_BUS_ERROR_NULL

    r = sd_bus_call_method(
        bus.bus,
        'org.bluez',
        '/org/bluez',
        "org.bluez.AgentManager1",
        'UnregisterAgent',
        &error,
        &msg,
        'o',
        '/org/btzen/Agent',
        NULL
    )
    sd_bus_error_free(&error);
    sd_bus_message_unref(msg);

    _sd_bus.check_call('unregister agent', r)

# vim: sw=4:et:ai



#!/usr/bin/env python3
#
# BTZen - library to asynchronously access Bluetooth devices.
#
# Copyright (C) 2015-2020 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#


"""
Build setup for the BTZen library.
"""

import ast
import sys
from setuptools import setup, find_packages, Extension

VERSION = ast.parse(
    next(l for l in open('btzen/__init__.py') if l.startswith('__version__'))
).body[0].value.s

try:
    from Cython.Build import cythonize
except:
    sys.exit(
        '\ncython is required, please install it with: pip install cython'
    )

setup(
    name='btzen',
    version=VERSION,
    author='Artur Wroblewski',
    author_email='wrobell@riseup.net',
    url='https://github.com/wrobell/btzen',
    description='BTZen - library to asynchronously access Bluetooth devices',
    setup_requires = ['setuptools_git >= 1.0',],
    classifiers=[
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Development Status :: 3 - Alpha',
        'Topic :: Software Development :: Libraries',
    ],
    ext_modules=cythonize([
        Extension('btzen._sd_bus', ['btzen/_sd_bus.pyx'], libraries=['systemd']),
        Extension('btzen._btzen', ['btzen/_btzen.pyx'], libraries=['systemd']),
        Extension('btzen._cm', ['btzen/_cm.pyx'], libraries=['systemd']),
    ]),
    packages=find_packages('.'),
    include_package_data=True,
    long_description=open('README').read(),
    long_description_content_type='text/x-rst',
)

# vim: sw=4:et:ai




#
# BTZen - library to asynchronously access Bluetooth devices.
#
# Copyright (C) 2015-2020 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# distutils: language = c
# cython: c_string_type=unicode, c_string_encoding=utf8, language_level=3str

from libc.stdio cimport perror
from libc.string cimport strerror
from libc.errno cimport errno

import asyncio
import logging

from ._sd_bus cimport *
from . import _sd_bus
from .error import *

logger = logging.getLogger(__name__)

FMT_RULE = """
type='signal',
sender='org.bluez',
interface='org.freedesktop.DBus.Properties',
member='PropertiesChanged',
path='{}',
arg0='{}'
"""

cdef class PropertyNotification:
    """
    Property notification based on asyncio queue class.

    The `put` method is used to add new value for a property. The `get`
    coroutine allows to retrieve property value in asynchronous manner.
    """
    cdef sd_bus_slot *slot
    cdef public object queues
    cdef public str path

    def __init__(self, path):
        self.queues = {}
        self.path = path

    def register(self, name):
        assert name not in self.queues
        assert self.slot is not NULL
        self.queues[name] = asyncio.Queue()

    def is_registered(self, name):
        return name in self.queues

    def put(self, name, value):
        assert name in self.queues
        assert self.slot is not NULL
        self.queues[name].put_nowait(value)

    async def get(self, name):
        assert name in self.queues
        assert self.slot is not NULL
        return (await self.queues[name].get())

    def size(self, name) -> int:
        return self.queues[name].qsize()

    def stop(self):
        self.queues.clear()
        sd_bus_slot_unref(self.slot)

cdef fmt_rule(iface, path):
    rule = FMT_RULE.format(path, iface)
    rule = rule.strip().replace('\n', '')
    return rule.encode()

cdef int task_cb_read(sd_bus_message *msg, void *user_data, sd_bus_error *ret_error) with gil:
    cdef object task = <object>user_data
    cdef const sd_bus_error *error = sd_bus_message_get_error(msg)
    cdef BusMessage bus_msg = BusMessage.__new__(BusMessage)
    bus_msg.c_obj = msg

    return _sd_bus.task_handle_message(bus_msg, task, DataReadError, 'ay')

async def bt_read(Bus bus, str path, timeout=0):
    assert bus is not None

    cdef sd_bus_message *msg = NULL
    cdef sd_bus_slot *slot = NULL

    task = asyncio.get_event_loop().create_future()

    r = sd_bus_message_new_method_call(
        bus.bus,
        &msg,
        'org.bluez',
        path.encode(),
        'org.bluez.GattCharacteristic1',
        'ReadValue'
    )
    _sd_bus.check_call('prepare read data from {}'.format(path), r)

    r = sd_bus_message_open_container(msg, 'a', '{sv}')
    _sd_bus.check_call('write data to {}'.format(path), r)

    r = sd_bus_message_close_container(msg)
    _sd_bus.check_call('write data to {}'.format(path), r)

    r = sd_bus_call_async(
        bus.bus, &slot,
        msg, task_cb_read,
        <void*>task,
        1e6 * timeout
    )
    _sd_bus.check_call('read data from {}'.format(path), r)

    try:
        return (await task)
    finally:
        sd_bus_message_unref(msg)
        sd_bus_slot_unref(slot)

cdef int task_cb_write(sd_bus_message *msg, void *user_data, sd_bus_error *ret_error) with gil:
    """
    Data write callback used by `bt_write` function.
    """
    cdef object task = <object>user_data
    cdef BusMessage bus_msg = BusMessage.__new__(BusMessage)
    bus_msg.c_obj = msg

    return _sd_bus.task_handle_message(bus_msg, task, DataWriteError, None)

async def bt_write(Bus bus, str path, bytes data):
    """
    Write data to Bluetooth device.

    :param bus: D-Bus reference.
    :param path: GATT characteristics path of the device.
    :param data: Data to write.
    """
    assert bus is not None

    cdef sd_bus_message *msg = NULL
    cdef sd_bus_slot *slot = NULL
    cdef char* buff = data

    task = asyncio.get_event_loop().create_future()
    try:
        r = sd_bus_message_new_method_call(
            bus.bus,
            &msg,
            'org.bluez',
            path.encode(),
            'org.bluez.GattCharacteristic1',
            'WriteValue'
        )
        _sd_bus.check_call('write data to {}'.format(path), r)

        r = sd_bus_message_append_array(msg, 'y', buff, len(data))
        _sd_bus.check_call('write data to {}'.format(path), r)

        r = sd_bus_message_open_container(msg, 'a', '{sv}')
        _sd_bus.check_call('write data to {}'.format(path), r)

        r = sd_bus_message_close_container(msg)
        _sd_bus.check_call('write data to {}'.format(path), r)

        r = sd_bus_call_async(bus.bus, &slot, msg, task_cb_write, <void*>task, 0)
        _sd_bus.check_call('write data to {}'.format(path), r)

        return (await task)
    finally:
        sd_bus_message_unref(msg)
        sd_bus_slot_unref(slot)

def bt_write_sync(Bus bus, str path, bytes data):
    """
    Write data to Bluetooth device.

    :param bus: D-Bus reference.
    :param path: GATT characteristics path of the device.
    :param data: Data to write.
    """
    assert bus is not None

    cdef sd_bus_message *msg = NULL
    cdef sd_bus_message *ret_msg = NULL
    cdef sd_bus_error error = SD_BUS_ERROR_NULL
    cdef char* buff = data

    r = sd_bus_message_new_method_call(
        bus.bus,
        &msg,
        'org.bluez',
        path.encode(),
        'org.bluez.GattCharacteristic1',
        'WriteValue'
    )
    _sd_bus.check_call('write data to {}'.format(path), r)

    r = sd_bus_message_append_array(msg, 'y', buff, len(data))
    _sd_bus.check_call('write data to {}'.format(path), r)

    r = sd_bus_message_open_container(msg, 'a', '{sv}')
    _sd_bus.check_call('write data to {}'.format(path), r)

    r = sd_bus_message_close_container(msg)
    _sd_bus.check_call('write data to {}'.format(path), r)

    r = sd_bus_call(bus.bus, msg, 0, &error, &ret_msg)
    _sd_bus.check_call('write data to {}'.format(path), r)

cdef int task_cb_property_monitor(sd_bus_message *msg, void *user_data, sd_bus_error *ret_error) with gil:
    cdef object cb = <object>user_data
    cdef const char *path
    cdef BusMessage bus_msg = BusMessage.__new__(BusMessage)

    path = sd_bus_message_get_path(msg)
    assert path == cb.path
    bus_msg.c_obj = msg

    # skip interface name
    _sd_bus.msg_skip(bus_msg, 's')

    for _ in _sd_bus.msg_container_dict(bus_msg, '{sv}'):
        name = _sd_bus.msg_read_value(bus_msg, 's')
        if cb.is_registered(name):
            value = _sd_bus.msg_read_value(bus_msg, 'v')
            cb.put(name, value)
        else:
            _sd_bus.msg_skip(bus_msg, 'v')

    return 0

def bt_property_monitor_start(Bus bus, str path, str iface):
    """
    Enable notification of value changes of Bluetooth device property.

    Property notification object is returned, which allows to register
    property names.

    :param bus: D-Bus reference.
    :param path: GATT characteristics path of the device.
    :param iface: Device interface.
    """
    assert bus is not None

    cdef sd_bus_slot *slot

    rule = fmt_rule(iface, path)
    data = PropertyNotification(path)

    r = sd_bus_add_match(
        bus.bus,
        &slot,
        rule,
        task_cb_property_monitor,
        <void*>data
    )
    _sd_bus.check_call('bus match rule', r)
    assert slot is not NULL

    data.slot = slot
    return data

async def bt_property(Bus bus, str path, str iface, str name, str type):
    assert bus is not None

    cdef sd_bus_message *msg = NULL
    cdef sd_bus_error error = SD_BUS_ERROR_NULL
    cdef BusMessage bus_msg = BusMessage.__new__(BusMessage)

    r = sd_bus_get_property(
        bus.bus,
        'org.bluez',
        path.encode(),
        iface.encode(),
        name.encode(),
        &error,
        &msg,
        type.encode()
    )
    _sd_bus.check_call('getting property {}'.format(name), r)

    bus_msg.c_obj = msg
    value =_sd_bus.msg_read_value(bus_msg, type)
    sd_bus_message_unref(msg)
    sd_bus_error_free(&error)

    return value

def bt_notify_start(Bus bus, str path):
    """
    Start monitoring value changes of a device identified by GATT
    characteristics path.

    :param bus: D-Bus reference.
    :param path: GATT characteristics path of the device.
    """
    assert bus is not None

    cdef sd_bus_message *msg = NULL
    cdef sd_bus_error error = SD_BUS_ERROR_NULL

    iface = 'org.bluez.GattCharacteristic1'

    r = sd_bus_call_method(
        bus.bus,
        'org.bluez',
        path.encode(),
        iface.encode(),
        'StartNotify',
        &error,
        &msg,
        NULL,
        NULL
    )
    sd_bus_error_free(&error);
    sd_bus_message_unref(msg);

    _sd_bus.check_call('start notification', r)

def bt_notify_stop(Bus bus, str path):
    """
    Stop monitoring value changes of a device identified by GATT
    characteristics path.

    :param bus: D-Bus reference.
    :param path: GATT characteristics path of the device.
    """
    assert bus is not None

    cdef sd_bus_message *msg = NULL
    cdef sd_bus_error error = SD_BUS_ERROR_NULL

    try:
        r = sd_bus_call_method(
            bus.bus,
            'org.bluez',
            path.encode(),
            'org.bluez.GattCharacteristic1',
            'StopNotify',
            &error,
            &msg,
            NULL,
            NULL
        )
        _sd_bus.check_call('stop notification', r)
    finally:
        sd_bus_error_free(&error)
        sd_bus_message_unref(msg)

def bt_process(Bus bus):
    """
    Process D-Bus events.
    """
    cdef int r
    assert bus is not None

    r = sd_bus_process(bus.bus, NULL)
    while r > 0:
        r = sd_bus_process(bus.bus, NULL)

def bt_characteristic(Bus bus, str prefix, str uuid):
    """
    Fetch Gatt Characteristic path relative to `prefix` and for specified
    UUID.
    """
    assert bus is not None

    cdef sd_bus_message *msg = NULL
    cdef sd_bus_error error = SD_BUS_ERROR_NULL
    cdef BusMessage bus_msg = BusMessage.__new__(BusMessage)

    try:
        r = sd_bus_call_method(
            bus.bus,
            'org.bluez',
            '/',
            'org.freedesktop.DBus.ObjectManager',
            'GetManagedObjects',
            &error,
            &msg,
            NULL
        )
        if r < 0:
            raise ConfigurationError(
                'Failed to get GATT characteristics paths: {}'
                .format(strerror(-r))
            )

        bus_msg.c_obj = msg

        path = _find_characteristic_path(bus_msg, prefix, uuid)

    finally:
        sd_bus_message_unref(msg)
        sd_bus_error_free(&error)

    return path

def _find_characteristic_path(BusMessage bus_msg, str prefix, str uuid):
    for _ in _sd_bus.msg_container_dict(bus_msg, '{oa{sa{sv}}}'):
        chr_path = _sd_bus.msg_read_value(bus_msg, 'o')

        if not chr_path.startswith(prefix):
             _sd_bus.msg_skip(bus_msg, 'a{sa{sv}}')
             continue

        for _ in _sd_bus.msg_container_dict(bus_msg, '{sa{sv}}'):
            iface = _sd_bus.msg_read_value(bus_msg, 's')

            if iface != 'org.bluez.GattCharacteristic1':
                _sd_bus.msg_skip(bus_msg, 'a{sv}')
                continue

            for _ in _sd_bus.msg_container_dict(bus_msg, '{sv}'):
                name = _sd_bus.msg_read_value(bus_msg, 's')
                if name == 'UUID':
                    value = _sd_bus.msg_read_value(bus_msg, 'v')
                    if uuid == value:
                        return chr_path
                else:
                    _sd_bus.msg_skip(bus_msg, 'v')
    return None

# vim: sw=4:et:ai
